/* $Id: heap3.c,v 1.1 2004/01/21 20:49:43 dw Exp $ */
/* a ternary heap */
/* node i has children 3i-1, 3i, 3i+1 */
/* and parent (i+1)/3 */


/* Author Damian Weber http://www-crypto.htw-saarland.de 
 * License: http://www.gnu.org/licenses/gpl-2.0.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/* heap sort for the a[0...n-1] */

/* heap elements are stored in a[1],...,a[size] */
struct heap3
{
	int *a;
	int size;
};

void heap3_init(struct heap3 *h, int max_size)
{
	h->size=0;
	h->a=(int *) malloc((max_size+1)*sizeof(int));
}

void heap3_destroy(struct heap3 *h)
{
	free(h->a);
}


/* insert e into the heap */
void heap3_insert(struct heap3 *h,int e)
{
	register int i,ip;
	int t;

	(h->size)++;
	i=h->size;
	h->a[h->size]=e;

	while (i>1)
	{
		ip=(i+1)/3;
		if (h->a[i]>h->a[ip])
		{
			t=h->a[i]; h->a[i]=h->a[ip]; h->a[ip]=t;
			i=ip;
		}
		else return;
	}
}

/* find maximum in the heap */
int heap3_find_max(struct heap3 *h)
{
	if (h->size>0) return h->a[1];
	printf("error: max on empty heap requested\n");
	return 0;
}

/* delete max from the heap */
void heap3_delete_max(struct heap3 *h)
{
	register int i,l,m,r,imax;
	int t;

	if (h->size<=0) 
	{
		printf("error: delete_max on empty heap requested\n");
		return;
	}


	h->a[1]=h->a[h->size];
	(h->size)--;

	i=1;
	while ((l=3*i-1)<=h->size)
	{
		m=l+1;
		r=l+2;
		imax=l;
		if (m<=h->size)
		{
			if (h->a[m]>h->a[imax])
				imax=m;
			if (r<=h->size && h->a[r]>h->a[imax])
				imax=r;
		}

		if (h->a[i]<h->a[imax])
		{
			t=h->a[i]; h->a[i]=h->a[imax]; h->a[imax]=t;
			i=imax;
		}
		else return; /* !!! exit from loop */
	}
}

/* sort array a[0],...,a[n-1] by heap sort */
void heap3_sort(int n,int *a)
{
	int i;
	struct heap3 h;

	heap3_init(&h,n);

	for (i=0;i<n;i++)
		heap3_insert(&h,a[i]);

	for (i=0;i<n;i++)
	{
		a[n-1-i]=heap3_find_max(&h);
		heap3_delete_max(&h);
	}
	heap3_destroy(&h);
}
